#ifdef USE_PRAGMA_IDENT_HDR
#pragma ident "@(#)psPromotionLAB.hpp	1.5 03/01/23 12:02:42 JVM"
#endif
/*
 * Copyright 2003 Sun Microsystems, Inc.  All rights reserved.
 * SUN PROPRIETARY/CONFIDENTIAL.  Use is subject to license terms.
 */

//
// PSPromotionLAB is a parallel scavenge promotion lab. This class acts very
// much like a MutableSpace. We couldn't embed a MutableSpace, though, as
// it has a considerable number of asserts and invariants that are violated.
//

class ObjectStartArray;

class PSPromotionLAB : public CHeapObj {
 protected:
  static const size_t filler_header_size;

  enum LabState {
    needs_flush,
    flushed,
    zero_size
  };

  HeapWord* _top;
  HeapWord* _bottom;
  HeapWord* _end;
  LabState _state;

  void set_top(HeapWord* value)    { _top = value; }
  void set_bottom(HeapWord* value) { _bottom = value; }
  void set_end(HeapWord* value)    { _end = value; }

  // The shared initialize code invokes this.
  debug_only(virtual bool lab_is_valid(MemRegion lab) { return false; });

  PSPromotionLAB() : _top(NULL), _bottom(NULL), _end(NULL) { }

 public:
  // Filling and flushing.
  void initialize(MemRegion lab);

  void flush();

  // Accessors
  HeapWord* bottom() const           { return _bottom; }
  HeapWord* end() const              { return _end;    }
  HeapWord* top() const              { return _top;    }

  bool is_flushed()                  { return _state == flushed; }

  bool unallocate_object(oop obj);

  // Returns a subregion containing all objects in this space.
  MemRegion used_region()            { return MemRegion(bottom(), top()); }

  // Boolean querries.
  bool is_empty() const              { return used() == 0; }
  bool not_empty() const             { return used() > 0; }
  bool contains(const void* p) const { return _bottom <= p && p < _end; }

  // Size computations.  Sizes are in bytes.
  size_t capacity() const            { return byte_size(bottom(), end()); }
  size_t used() const                { return byte_size(bottom(), top()); }
  size_t free() const                { return byte_size(top(),    end()); }
};

class PSYoungPromotionLAB : public PSPromotionLAB {
 public:
  PSYoungPromotionLAB() { }

  // Not MT safe
  HeapWord* allocate(size_t size) {
    // Can't assert this, when young fills, we keep the LAB around, but flushed.
    // assert(_state != flushed, "Sanity");
    HeapWord* obj = top();
    HeapWord* new_top = obj + size;
    // The 'new_top>obj' check is needed to detect overflow of obj+size.
    if (new_top > obj && new_top <= end()) {
      set_top(new_top);
      assert(is_object_aligned((intptr_t)obj) && is_object_aligned((intptr_t)new_top),
             "checking alignment");
      return obj;
    }

    return NULL;
  }

  debug_only(virtual bool lab_is_valid(MemRegion lab));
};

class PSOldPromotionLAB : public PSPromotionLAB {
 private:
  ObjectStartArray* _start_array;

 public:
  PSOldPromotionLAB() : _start_array(NULL) { }
  PSOldPromotionLAB(ObjectStartArray* start_array) : _start_array(start_array) { }

  void set_start_array(ObjectStartArray* start_array) { _start_array = start_array; }

  // Not MT safe
  HeapWord* allocate(size_t size) {
    // Cannot test for this now that we're doing promotion failures
    // assert(_state != flushed, "Sanity");
    assert(_start_array != NULL, "Sanity");
    HeapWord* obj = top();
    HeapWord* new_top = obj + size;
    // The 'new_top>obj' check is needed to detect overflow of obj+size.
    if (new_top > obj && new_top <= end()) {
      set_top(new_top);
      assert(is_object_aligned((intptr_t)obj) && is_object_aligned((intptr_t)new_top),
             "checking alignment");
      _start_array->allocate_block(obj);
      return obj;
    }

    return NULL;
  }

  debug_only(virtual bool lab_is_valid(MemRegion lab));
};


