#ifdef USE_PRAGMA_IDENT_HDR
#pragma ident "@(#)genCollectedHeap.hpp	1.56 03/01/23 12:08:20 JVM"
#endif
/*
 * Copyright 2003 Sun Microsystems, Inc.  All rights reserved.
 * SUN PROPRIETARY/CONFIDENTIAL.  Use is subject to license terms.
 */

class SubTasksDone;

// A "GenCollectedHeap" is a SharedHeap that uses generational
// collection.  It is represented with a sequence of Generation's.
class GenCollectedHeap : public SharedHeap {
  friend class TwoGenerationCollectorPolicy;
  friend class Generation;
  friend class DefNewGeneration;
  friend class TenuredGeneration;
  friend class TrainGeneration;
  friend class ConcurrentMarkSweepGeneration;
  friend class CMSCollector;
  friend class GenMarkSweep;
  friend class VM_GenCollectForAllocation;
  friend class VM_GenCollectFull;
  friend class VM_GC_HeapInspection;
  friend class HeapInspection;
  friend class GCCauseSetter;
  friend class VMStructs;
public:
  enum SomeConstants {
    max_gens = 10
  };

private:
  // Fields:
  static GenCollectedHeap* _gch;

  int _n_gens;
  Generation* _gens[max_gens];
  GenerationSpec** _gen_specs;

  // If a generation would bail out of an incremental collection, 
  // it sets this flag.  If the flag is set, satisfy_failed_allocation 
  // will attempt allocating in all generations before doing a full GC.
  bool _incremental_collection_will_fail;

  // Data structure for claiming the (potentially) parallel tasks in
  // (gen-specific) strong roots processing.
  SubTasksDone* _gen_process_strong_tasks;

  // In block contents verification, the number of header words to skip
  NOT_PRODUCT(static size_t _skip_header_HeapWords;)

protected:
  // Directs each generation up to and including "collectedGen" to recompute
  // its desired size.
  void compute_new_generation_sizes(int collectedGen);

  // Helper functions for allocation
  HeapWord* attempt_allocation(size_t size,
                               bool   is_large_noref,
                               bool   is_tlab,
                               bool   first_only);

  // Helper function for two callbacks below.
  // Considers collection of the first max_level+1 generations.
  void do_collection(bool   full,
                     bool   clear_all_soft_refs,
                     size_t size,
                     bool   is_large_noref,
                     bool   is_tlab,
		     int    max_level,
		     bool&  notify_ref_lock);

  // Callback from VM_GenCollectForAllocation operation.
  // This function does everything necessary/possible to satisfy an
  // allocation request that failed in the youngest generation that should
  // have handled it (including collection, expansion, etc.)  Sets
  // "notify_ref_lock" to true if it enqueues any weak references.
  HeapWord* satisfy_failed_allocation(size_t size,
                                      bool   is_large_noref,
                                      bool   is_tlab, 
				      bool&  notify_ref_lock);

  // Callback from VM_GenCollectFull operation.
  // Perform a full collection of the first max_level+1 generations.
  void do_full_collection(bool  clear_all_soft_refs,
                          int   max_level, 
                          bool& notify_ref_lock);

public:
  GenCollectedHeap(CollectorPolicy *policy);

  // Does not return if initialization fails.
  void initialize();

  // Does operations required after initialization has been done.
  void post_initialize();

  // Initialize ("weak") refs processing support
  virtual void ref_processing_init();

  virtual CollectedHeap::Name kind() { return CollectedHeap::GenCollectedHeap; }

  size_t capacity() const;
  size_t used() const;

  size_t max_capacity() const;

  HeapWord* mem_allocate(size_t size,
                         bool   is_large_noref,
                         bool   is_tlab);

  // We may support a shared contiguous allocation area, if the youngest
  // generation does.
  bool supports_inline_contig_alloc() const;
  HeapWord** top_addr() const;
  HeapWord** end_addr() const;

  // Return an estimate of the maximum allocation that could be performed
  // without triggering any collection activity.  In a generational
  // collector, for example, this is probably the largest allocation that
  // could be supported in the youngest generation.  It is "unsafe" because
  // no locks are taken; the result should be treated as an approximation,
  // not a guarantee.
  size_t unsafe_max_alloc();

  // Perform a full collection of the heap; intended for use in implementing
  // "System.gc". This implies as full a collection as the CollectedHeap
  // supports. Caller does not hold the Heap_lock on entry.
  void collect(GCCause::Cause cause);

  // The same as above but assume that the caller holds the Heap_lock.
  void collect_locked(GCCause::Cause cause);

  // Perform a full collection of the first max_level+1 generations.
  // Mostly used for testing purposes. Caller does not hold the Heap_lock on entry.
  void collect(GCCause::Cause cause, int max_level);

  // Returns "TRUE" iff "p" points into the allocated area of the heap.
  bool is_in(const void* p) const;

  // Returns "TRUE" iff "p" points into the youngest generation.
  bool is_in_youngest(void* p);

  // Iteration functions.
  void oop_iterate(OopClosure* cl);
  void oop_iterate(MemRegion mr, OopClosure* cl);
  void object_iterate(ObjectClosure* cl);
  void object_iterate_since_last_GC(ObjectClosure* cl);
  Space* space_containing(const void* addr) const;

  // A CollectedHeap is divided into a dense sequence of "blocks"; that is,
  // each address in the (reserved) heap is a member of exactly
  // one block.  The defining characteristic of a block is that it is
  // possible to find its size, and thus to progress forward to the next
  // block.  (Blocks may be of different sizes.)  Thus, blocks may
  // represent Java objects, or they might be free blocks in a
  // free-list-based heap (or subheap), as long as the two kinds are 
  // distinguishable and the size of each is determinable.

  // Returns the address of the start of the "block" that contains the
  // address "addr".  We say "blocks" instead of "object" since some heaps
  // may not pack objects densely; a chunk may either be an object or a
  // non-object. 
  virtual HeapWord* block_start(const void* addr) const;

  // Requires "addr" to be the start of a chunk, and returns its size.
  // "addr + size" is required to be the start of a new chunk, or the end
  // of the active area of the heap.
  virtual size_t block_size(const HeapWord* addr) const;

  // Requires "addr" to be the start of a block, and returns "TRUE" iff
  // the block is an object.
  virtual bool block_is_obj(const HeapWord* addr) const;

  // Section on TLAB's.
  virtual bool supports_tlab_allocation() const;
  virtual size_t tlab_capacity() const;
  virtual HeapWord* allocate_new_tlab(size_t size);

  // The "requestor" generation is performing some garbage collection 
  // action for which it would be useful to have scratch space.  The
  // requestor promises to allocate no more than "max_alloc_words" in any
  // older generation (via promotion say.)   Any blocks of space that can
  // be provided are returned as a list of ScratchBlocks, sorted by
  // decreasing size.
  ScratchBlock* gather_scratch(Generation* requestor, size_t max_alloc_words);

  size_t large_typearray_limit();

  // Iterate through each generation, giving them a chance
  // to examine a modified oop, "p", and take appropriate
  // actions, such as updating internal generational data.
  void examine_modified_oop(oop* p) {
    for (int i = 0; i < _n_gens; i++) {
      _gens[i]->examine_modified_oop(p);
    }
    perm_gen()->examine_modified_oop(p);
  }

  // Ensure parseability: override
  virtual void ensure_parseability();

  // Time in ms since the longest time a collector ran in
  // in any generation.
  virtual jlong millis_since_last_gc();

  // Update "time of last gc" for all constituent generations
  // to "now".
  void update_time_of_last_gc(jlong now) {
    for (int i = 0; i < _n_gens; i++) {
      _gens[i]->update_time_of_last_gc(now);
    }
    perm_gen()->update_time_of_last_gc(now);
  }

  // Override.
  bool no_gc_in_progress() { return !is_gc_active(); }

  // Override.
  void prepare_for_verify() PRODUCT_RETURN;

  // Override.
  void verify(bool allow_dirty, bool silent) PRODUCT_RETURN;

  // Override.
  void print() const;
  void print_on(outputStream* st) const;

  // PrintGC, PrintGCDetails support
  void print_heap_change(size_t prev_used) const;
  void print_perm_heap_change(size_t perm_prev_used) const;

  int addr_to_arena_id(void* addr);

  // The functions below are helper functions that a subclass of
  // "CollectedHeap" can use in the implementation of its virtual
  // functions.

  class GenClosure : public StackObj {
   public:
    virtual void do_generation(Generation* gen) = 0;
  };

  // Apply "cl.do_generation" to all generations in the heap (not including
  // the permanent generation).  If "old_to_young" determines the order.
  void generation_iterate(GenClosure* cl, bool old_to_young);

  void space_iterate(SpaceClosure* cl);

  // Return the generation before "gen", or else NULL.
  Generation* prev_gen(Generation* gen) const {
    int l = gen->level();
    if (l == 0) return NULL;
    else return _gens[l-1];
  }

  // Return the generation after "gen", or else NULL.
  Generation* next_gen(Generation* gen) const {
    int l = gen->level() + 1;
    if (l == _n_gens) return NULL;
    else return _gens[l];
  }

  Generation* get_gen(int i) {
    if (i >= 0 && i < _n_gens)
      return _gens[i];
    else
      return NULL;
  }

  int n_gens() const {
    assert(_n_gens == collector_policy()->number_of_generations(), "Sanity");
    return _n_gens;
  }

  // Convenience function to be used in situations where the heap type can be
  // asserted to be this type.
  static GenCollectedHeap* heap();

  enum ClassScanningOption {
    CSO_None,
    CSO_AllClasses,
    CSO_SystemClasses
  };

  void set_par_threads(int t);
  

  // Invoke the "do_oop" method of one of the closures "not_older_gens"
  // or "older_gens" on root locations for the generation at
  // "level".  (The "older_gens" closure is used for scanning references
  // from older generations; "not_older_gens" is used everywhere else.)
  // If "younger_gens_as_roots" is false, younger generations are
  // not scanned as roots; in this case, the caller must be arranging to
  // scan the younger generations itself.  (For example, a generation might
  // explicitly mark reachable objects in younger generations, to avoid
  // excess storage retention.)  If "collecting_perm_gen" is false, then
  // roots that may only contain references to permGen objects are not
  // scanned.  The "cso" argument determines which entries of the
  // SystemDictionary the closure is applied to.  "CSO_none" does none;
  // "CSO_AllClasses" applies the closure to all entries for "system"
  // classes and loaders.
  void process_strong_roots(int level, bool younger_gens_as_roots,
			    bool collecting_perm_gen,
			    ClassScanningOption cso,
			    OopsInGenClosure* older_gens,
			    OopsInGenClosure* not_older_gens);

  // Apply "blk" to all the weak roots of the system.  These include
  // JNI weak roots, the code cache, system dictionary, symbol table,
  // string table, and referents of reachable weak refs. 
  void process_weak_roots(OopClosure* root_closure,
			  OopClosure* non_root_closure);

  // Set the saved marks of generations, if that makes sense.
  // In particular, if any generation might iterate over the oops
  // in other generations, it should call this method.
  void save_marks();

  // Apply "cur->do_oop" or "older->do_oop" to all the oops in objects
  // allocated since the last call to save_marks in generations at or above
  // "level" (including the permanent generation.)  The "cur" closure is
  // applied to references in the generation at "level", and the "older"
  // closure to older (and permanent) generations.
#define GCH_SINCE_SAVE_MARKS_ITERATE_DECL(OopClosureType, nv_suffix)	\
  void oop_since_save_marks_iterate(int level,				\
			            OopClosureType* cur,		\
				    OopClosureType* older);

  ALL_SINCE_SAVE_MARKS_CLOSURES(GCH_SINCE_SAVE_MARKS_ITERATE_DECL)

#undef GCH_SINCE_SAVE_MARKS_ITERATE_DECL

  // Returns "true" iff no allocations have occurred in any generation at
  // "level" or above (including the permanent generation) since the last
  // call to "save_marks".
  bool no_allocs_since_save_marks(int level);

  // If a generation bails out of an incremental collection, 
  // it sets this flag.
  bool incremental_collection_will_fail() {
    return _incremental_collection_will_fail;
  }
  void set_incremental_collection_will_fail() {
    _incremental_collection_will_fail = true;
  }
  void clear_incremental_collection_will_fail() {
    _incremental_collection_will_fail = false;
  }

  // Promotion of obj into gen failed.  Try to promote obj to higher non-perm
  // gens in ascending order; return the new location of obj if successful.
  // Otherwise, try expand-and-allocate for obj in each generation starting at
  // gen; return the new location of obj if successful.  Otherwise, return NULL.
  oop handle_failed_promotion(Generation* gen,
			      oop obj,
			      size_t obj_size,
			      oop* ref);

private:
  // Accessor for memory state verification support
  NOT_PRODUCT(
    static size_t skip_header_HeapWords() { return _skip_header_HeapWords; }
  )

  // Override
  void check_for_non_bad_heap_word_value(HeapWord* addr,
    size_t size) PRODUCT_RETURN;

  // For use by mark-sweep.  As implemented, mark-sweep-compact is global
  // in an essential way: compaction is performed across generations, by
  // iterating over spaces.
  void prepare_for_compaction();

  // Perform a full collection of the first max_level+1 generations.
  // This is the low level interface used by the public versions of
  // collect() and collect_locked(). Caller holds the Heap_lock on entry.
  void collect_locked(GCCause::Cause cause, int max_level);

protected:
  virtual void gc_prologue(bool full);
  virtual void gc_epilogue(bool full);
  
};

// Class to set and reset the GC cause for a GenCollectedHeap.

class GCCauseSetter : StackObj {
  GenCollectedHeap* _heap;
  GCCause::Cause _previous_cause;
 public:
  GCCauseSetter(GenCollectedHeap* heap, GCCause::Cause cause) {
    _heap = heap;
    _previous_cause = _heap->gc_cause();
    _heap->set_gc_cause(cause);
  }
  ~GCCauseSetter() {
    _heap->set_gc_cause(_previous_cause);
  }
};

