/*
 * @(#)RepaintArea.java	1.17 03/04/14
 *
 * Copyright 2003 Sun Microsystems, Inc. All rights reserved.
 * SUN PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */

package sun.awt;

import java.awt.Component;
import java.awt.Graphics;
import java.awt.Point;
import java.awt.Rectangle; 
import java.awt.event.PaintEvent;

/**
 * The <code>RepaintArea</code> is a geometric construct created for the 
 * purpose of holding the geometry of several coalesced paint events.  
 * This geometry is accessed synchronously, although it is written such 
 * that painting may still be executed asynchronously.  
 *
 * @author 	Eric Hawkes
 * @version 	1.17 04/14/03
 * @since 	1.3
 */
public class RepaintArea {

    /** 
     * Maximum ratio of bounding rectangle to benefit for which 
     * both the vertical and horizontal unions are repainted. 
     * For smaller ratios the whole bounding rectangle is repainted. 
     * @see #paint 
     */ 
    private static final int MAX_BENEFIT_RATIO = 4; 
     
    private static final int HORIZONTAL = 0;
    private static final int VERTICAL = 1;
    private static final int UPDATE = 2;
 
    private Rectangle paintRects[] = new Rectangle[UPDATE + 1];

 
    /**
     * Constructs a new <code>RepaintArea</code> 
     * @since 	1.3
     */ 
    public RepaintArea() {
    }

    /**
     * Constructs a new <code>RepaintArea</code> initialized to match 
     * the values of the specified RepaintArea.  
     *
     * @param   ra  the <code>RepaintArea</code> from which to copy initial 
     *              values to a newly constructed RepaintArea
     * @since 	1.3
     */ 
    private RepaintArea(RepaintArea ra) {
        // This constructor is private because it should only be called 
	// from the cloneAndReset method
        for (int i = 0; i < paintRects.length; i++) {
            paintRects[i] = ra.paintRects[i];
        }
    }

    /**
     * Adds a <code>Rectangle</code> to this <code>RepaintArea</code>. 
     * PAINT Rectangles are divided into mostly vertical and mostly horizontal.
     * Each group is unioned together.
     * UPDATE Rectangles are unioned.
     *
     * @param   r   the specified <code>Rectangle</code>
     * @param   id  possible values PaintEvent.UPDATE or PaintEvent.PAINT
     * @since 	1.3
     */
    public synchronized void add(Rectangle r, int id) {
        // Make sure this new rectangle has positive dimensions 
        if (r.isEmpty()) {
	    return;
	}
        int addTo = UPDATE;
        if (id == PaintEvent.PAINT) {
            addTo = (r.width > r.height) ? HORIZONTAL : VERTICAL;
        }
        if (paintRects[addTo] != null) {
            paintRects[addTo].add(r);
        } else {
            paintRects[addTo] = new Rectangle(r);
        }
    }


    /**
     * Creates a new <code>RepaintArea</code> with the same geometry as this 
     * RepaintArea, then removes all of the geometry from this 
     * RepaintArea and restores it to an empty RepaintArea.  
     *
     * @return	ra a new <code>RepaintArea</code> having the same geometry as 
     *          this RepaintArea.  
     * @since 	1.3
     */
    private synchronized RepaintArea cloneAndReset() { 
        RepaintArea ra = new RepaintArea(this);
        for (int i = 0; i < paintRects.length; i++) {
            paintRects[i] = null;
        }
        return ra;
    }

    public boolean isEmpty() {
        for (int i = 0; i < paintRects.length; i++) {
            if (paintRects[i] != null) {
                return false;
            }
        }
        return true;
    }    

    /**
     * Invokes paint and update on target Component with optimal
     * rectangular clip region.
     * If PAINT bounding rectangle is less than
     * MAX_BENEFIT_RATIO times the benefit, then the vertical and horizontal unions are
     * painted separately.  Otherwise the entire bounding rectangle is painted.
     *
     * @param   target Component to <code>paint</code> or <code>update</code>
     * @since   1.4
     */
    public void paint(Object target, boolean shouldClearRectBeforePaint) {
	Component comp = (Component)target; 

	if (isEmpty()) { 
	    return; 
        } 

        if (!comp.isVisible()) {
            return;
        }

        RepaintArea ra = this.cloneAndReset(); 

        if (!subtract(ra.paintRects[VERTICAL], ra.paintRects[HORIZONTAL])) {
            subtract(ra.paintRects[HORIZONTAL], ra.paintRects[VERTICAL]); 
        }

        if (ra.paintRects[HORIZONTAL] != null && ra.paintRects[VERTICAL] != null) { 
            Rectangle paintRect = ra.paintRects[HORIZONTAL].union(ra.paintRects[VERTICAL]);
            int square = paintRect.width * paintRect.height; 
            int benefit = square - ra.paintRects[HORIZONTAL].width 
                * ra.paintRects[HORIZONTAL].height - ra.paintRects[VERTICAL].width 
                * ra.paintRects[VERTICAL].height; 
            // if benefit is comparable with bounding box 
            if (MAX_BENEFIT_RATIO * benefit < square) { 
                ra.paintRects[HORIZONTAL] = paintRect;
                ra.paintRects[VERTICAL] = null;
            }
        } 
        for (int i = 0; i < paintRects.length; i++) {
            if (ra.paintRects[i] != null 
                && !ra.paintRects[i].isEmpty()) 
            {
                // Should use separate Graphics for each paint() call,
                // since paint() can change Graphics state for next call. 
                Graphics g = comp.getGraphics();
                if (g != null) {
                    try {
                        g.setClip(ra.paintRects[i]);
                        if (i == UPDATE) {
                            comp.update(g);
                        } else {
                            if (shouldClearRectBeforePaint) {
                                g.clearRect( ra.paintRects[i].x,
                                             ra.paintRects[i].y,
                                             ra.paintRects[i].width,
                                             ra.paintRects[i].height);
                            }
                            comp.paint(g);
                        }
                    } finally {
                        g.dispose();
                    }
                }
            } 
        }
    }

    /**
     * Subtracts subtr from rect. If the result is rectangle
     * changes rect and returns true. Otherwise false.
     */
    static boolean subtract(Rectangle rect, Rectangle subtr) {
        if (rect == null || subtr == null) {
            return true;
        }
        Rectangle common = rect.intersection(subtr);
        if (common.isEmpty()) {
            return true;
        }
        if (rect.x == common.x && rect.y == common.y) {
            if (rect.width == common.width) {
                rect.y += common.height;
                rect.height -= common.height;
                return true;
            } else
            if (rect.height == common.height) {
                rect.x += common.width;
                rect.width -= common.width;
                return true;
            } 
        } else       
        if (rect.x + rect.width == common.x + common.width 
            && rect.y + rect.height == common.y + common.height) 
        {
            if (rect.width == common.width) {
                rect.height -= common.height;
                return true;
            } else
            if (rect.height == common.height) {
                rect.width -= common.width;
                return true;
            } 
        }
        return false;
    }
}

