/*
 * @(#)GeneralNames.java	1.16 03/01/23
 *
 * Copyright 2003 Sun Microsystems, Inc. All rights reserved.
 * SUN PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */

package sun.security.x509;

import java.util.*;
import java.io.IOException;

import sun.security.util.*;

/**
 * This object class represents the GeneralNames type required in
 * X509 certificates.  
 * <p>The ASN.1 syntax for this is:
 * <pre>
 * GeneralNames ::= SEQUENCE SIZE (1..MAX) OF GeneralName
 * </pre>
 *
 * @author Amit Kapoor
 * @author Hemma Prafullchandra
 *
 * @version 1.16, 01/23/03
 */
public class GeneralNames {
    
    private final List names;

    /**
     * Create the GeneralNames, decoding from the passed DerValue.
     *
     * @param derVal the DerValue to construct the GeneralNames from.
     * @exception IOException on error.
     */
    public GeneralNames(DerValue derVal) throws IOException {
	this();
        if (derVal.tag != DerValue.tag_Sequence) {
            throw new IOException("Invalid encoding for GeneralNames.");
        }
        if (derVal.data.available() == 0) {
            throw new IOException("No data available in "
                                      + "passed DER encoded value.");
        }
        // Decode all the GeneralName's
        while (derVal.data.available() != 0) {
            DerValue encName = derVal.data.getDerValue();

            GeneralName name = new GeneralName(encName);
            add(name);
        }
    }

    /**
     * The default constructor for this class.
     */
    public GeneralNames() {
	names = new ArrayList();
    }
    
    public GeneralNames add(GeneralName name) {
	if (name == null) {
	    throw new NullPointerException();
	}
	names.add(name);
	return this;
    }
    
    public GeneralName get(int index) {
	return (GeneralName)names.get(index);
    }
    
    public boolean isEmpty() {
	return names.isEmpty();
    }
    
    public int size() {
	return names.size();
    }
    
    public Iterator iterator() {
	return names.iterator();
    }

    /**
     * Write the extension to the DerOutputStream.
     *
     * @param out the DerOutputStream to write the extension to.
     * @exception IOException on error.
     */
    public void encode(DerOutputStream out) throws IOException {
        if (isEmpty()) {
	    return;
	}

        DerOutputStream temp = new DerOutputStream();
	for (Iterator t = iterator(); t.hasNext(); ) {
	    GeneralName gn = (GeneralName)t.next();
            gn.encode(temp);
        }
        out.write(DerValue.tag_Sequence, temp);
    }

    /**
     * compare this GeneralNames to other object for equality
     *
     * @returns true iff this equals other
     */
    public boolean equals(Object obj) {
	if (this == obj) {
	    return true;
	}
	if (obj instanceof GeneralNames == false) {
	    return false;
	}
	GeneralNames other = (GeneralNames)obj;
	return this.names.equals(other.names);
    }
    
    public int hashCode() {
	return names.hashCode();
    }
    
    public String toString() {
	return names.toString();
    }

}
