/**
 * @(#)DocLocale.java	1.19 03/01/23
 *
 * Copyright 2003 Sun Microsystems, Inc. All rights reserved.
 * SUN PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */
package com.sun.tools.javadoc;
import java.util.Locale;

import java.text.Collator;

import java.text.BreakIterator;


/**
 * This class holds the information about locales.
 *
 * @since JDK1.4
 * @author Robert Field
 */
class DocLocale {

    /**
     * The locale name will be set by Main, if option is provided on the
     * command line.
     */
    final String localeName;

    /**
     * The locale to be used. If user doesen't provide this,
     * then set it to default locale value.
     */
    final Locale locale;

    /**
     * The collator for this application. This is to take care of Locale
     * Specific or Natural Language Text sorting.
     */
    final Collator collator;

    /**
     * Enclosing DocEnv
     */
    private final DocEnv docenv;

    /**
     * Sentence instance from the BreakIterator.
     */
    private final BreakIterator sentenceBreaker;

    /**
     * True is we should use <code>BreakIterator</code>
     * to compute first sentence.
     */
    private boolean useBreakIterator = false;

    /**
     * True if the long version of breakiterator warning has been
     * printed.
     */
    private boolean printedLongBreakIteratorWarning = false;

    /**
     * The HTML sentence terminators.
     */
    static final String[] sentenceTerminators = {"<p>", "</p>", "<h1>", "<h2>", "<h3>",
    "<h4>", "<h5>", "<h6>", "</h1>", "</h2>", "</h3>", "</h4>", "</h5>",
    "</h6>", "<hr>", "<pre>", "</pre>"};

    /**
     * Constructor
     */
    DocLocale(DocEnv docenv, String localeName, boolean useBreakIterator) {
        super();
        this.docenv = docenv;
        this.localeName = localeName;
        this.useBreakIterator = useBreakIterator;
        locale = getLocale();
        if (locale == null) {
            docenv.exit();
        } else {
            Locale.setDefault(locale);
        }
        collator = Collator.getInstance(locale);
        sentenceBreaker = BreakIterator.getSentenceInstance(locale);
    }

    /**
      * Get the locale if specified on the command line
      * else return null and if locale option is not used
      * then return default locale.
      */
    private Locale getLocale() {
        Locale userlocale = null;
        if (localeName.length() > 0) {
            int firstuscore = localeName.indexOf('_');
            int seconduscore = -1;
            String language = null;
            String country = null;
            String variant = null;
            if (firstuscore == 2) {
                language = localeName.substring(0, firstuscore);
                seconduscore = localeName.indexOf('_', firstuscore + 1);
                if (seconduscore > 0) {
                    if (seconduscore != firstuscore + 3 ||
                            localeName.length() <= seconduscore + 1) {
                        docenv.error(null, "main.malformed_locale_name", localeName);
                        return null;
                    }
                    country = localeName.substring(firstuscore + 1, seconduscore);
                    variant = localeName.substring(seconduscore + 1);
                } else if (localeName.length() == firstuscore + 3) {
                    country = localeName.substring(firstuscore + 1);
                } else {
                    docenv.error(null, "main.malformed_locale_name", localeName);
                    return null;
                }
            } else if (firstuscore == -1 && localeName.length() == 2) {
                language = localeName;
            } else {
                docenv.error(null, "main.malformed_locale_name", localeName);
                return null;
            }
            userlocale = searchLocale(language, country, variant);
            if (userlocale == null) {
                docenv.error(null, "main.illegal_locale_name", localeName);
                return null;
            } else {
                return userlocale;
            }
        } else {
            return Locale.getDefault();
        }
    }

    /**
      * Search the locale for specified language, specified country and
      * specified variant.
      */
    private Locale searchLocale(String language, String country, String variant) {
        Locale[] locales = Locale.getAvailableLocales();
        for (int i = 0; i < locales.length; i++) {
            if (locales[i].getLanguage().equals(language) && (country == null ||
                    locales[i].getCountry().equals(country)) &&
                    (variant == null || locales[i].getVariant().equals(variant))) {
                return locales[i];
            }
        }
        return null;
    }

    String localeSpecificFirstSentence(DocImpl doc, String s) {
        if (s == null || s.length() == 0) {
            return "";
        }
        int index = s.indexOf("-->");
        if (s.trim().startsWith("<!--") && index != -1) {
            return localeSpecificFirstSentence(doc,
                    s.substring(index + 3, s.length()));
        }
        sentenceBreaker.setText(s.replace('\n', ' '));
        int start = sentenceBreaker.first();
        int end = sentenceBreaker.next();
        String result = s.substring(start, end);
        if (locale.getLanguage().equals("en") && !useBreakIterator) {
            String oldResult = englishLanguageFirstSentence(s);
            if (!result.trim().equals(oldResult.trim())) {
                if (!printedLongBreakIteratorWarning) {
                    docenv.warning(doc, "javadoc.BreakIterator_Error_Long_0");
                    docenv.notice("javadoc.BreakIterator_Error_Long_1", oldResult);
                    docenv.notice("javadoc.BreakIterator_Error_Long_2");
                    docenv.notice("javadoc.BreakIterator_Error_Long_3");
                    docenv.notice("javadoc.BreakIterator_Error_Long_4", result);
                    docenv.notice("javadoc.BreakIterator_Error_Long_5");
                    docenv.notice("javadoc.BreakIterator_Error_Long_6");
                    printedLongBreakIteratorWarning = true;
                } else {
                    docenv.warning(doc, "javadoc.BreakIterator_Error_Short_0");
                    docenv.notice("javadoc.BreakIterator_Error_Short_1", oldResult);
                    docenv.notice("javadoc.BreakIterator_Error_Short_2");
                    docenv.notice("javadoc.BreakIterator_Error_Short_3", result);
                }
            }
            return oldResult;
        }
        return result;
    }

    /**
      * Return the first sentence of a string, where a sentence ends
      * with a period followed be white space.
      */
    private String englishLanguageFirstSentence(String s) {
        if (s == null) {
            return null;
        }
        int len = s.length();
        boolean period = false;
        for (int i = 0; i < len; i++) {
            switch (s.charAt(i)) {
            case '.':
                period = true;
                break;

            case ' ':

            case '\t':

            case '\n':

            case '\r':

            case '\f':
                if (period) {
                    return s.substring(0, i);
                }
                break;

            case '<':
                if (i > 0) {
                    if (htmlSentenceTerminatorFound(s, i)) {
                        return s.substring(0, i);
                    }
                }
                break;

            default:
                period = false;

            }
        }
        return s;
    }

    /**
      * Find out if there is any HTML tag in the given string. If found
      * return true else return false.
      */
    private boolean htmlSentenceTerminatorFound(String str, int index) {
        for (int i = 0; i < sentenceTerminators.length; i++) {
            String terminator = sentenceTerminators[i];
            if (str.regionMatches(true, index, terminator, 0, terminator.length())) {
                return true;
            }
        }
        return false;
    }
}
