/*
 * @(#)PlatformMidi.h	1.14 03/01/23
 *
 * Copyright 2003 Sun Microsystems, Inc. All rights reserved.
 * SUN PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */

/*
 * $$kk: 06.23.99
 * This file contains header information for platform midi support.
 * This does not interface with the Beatnik engine at all.
 */
#ifndef PLATFORM_MIDI_INCLUDED
#define PLATFORM_MIDI_INCLUDED

// includes for types
#include "engine/X_API.h"

#include "Configure.h"				// put flags for debug msgs etc. here
#include "Utilities.h"

/* *********************** MIDI TYPES (for all platforms) ******************************* */

// MIDI message types
typedef enum {
    SHORT_MESSAGE = 0,
    LONG_MESSAGE = 1
} MidiMessageType;

// MIDI message object
typedef struct tag_MidiMessage {
    INT64 timestamp;  // in microseconds
    INT32 locked;     // TRUE when event is currently being read
    MidiMessageType type;
    union {
	struct {
	    // platform-endianness packed message:
	    // status | data1<<8 | data2<<16
	    UINT32 packedMsg;
	} s; // short message
	struct {
	    UINT32  size;
	    // this buffer is read only. It must not be freed.
	    UBYTE* data;
	    INT32 index; // sysex buffer number
	} l; // long message
    } data;
} MidiMessage;

/*
 * Native MIDI message circular buffer
 */
typedef struct tag_MidiQueue {
    void* lock;
    INT32 size;
    INT32 capacity;
    INT32 readIndex;
    INT32 writeIndex;
    MidiMessage queue[1];
} MidiMessageQueue;

// device handle, to be created and filled in MIDI_IN_OpenDevice() and MIDI_OUT_OpenDevice()
typedef struct tag_MidiDeviceHandle {
    void* deviceHandle;      // handle to the device
    MidiMessageQueue* queue; // may be NULL if no queue is used
    void* longBuffers;       // contains platform-specific data for long buffers, e.g. list of MIDIHDR
    void* platformData;      // contains platform specific data, e.g. an Event object
    INT32 isWaiting;         // if TRUE, then waiting for new events
} MidiDeviceHandle;


#if (USE_PLATFORM_MIDI_IN == TRUE) || (USE_PLATFORM_MIDI_OUT == TRUE)

/*
 * Native Locking support
 */
void* MIDI_CreateLock();
void MIDI_DestroyLock(void* lock);

/* Blocks until this lock can be gotten.
 * Nop if lock is NULL */
void MIDI_Lock(void* lock);

/* Releases this lock */
void MIDI_Unlock(void* lock);

MidiMessageQueue* MIDI_CreateQueue(int capacity);
void MIDI_DestroyQueue(MidiMessageQueue* queue);
// if overwrite is true, oldest messages will be overwritten when the queue is full
// returns true, if message has been added
int MIDI_QueueAddShort(MidiMessageQueue* queue, UINT32 packedMsg, INT64 timestamp, int overwrite);
int MIDI_QueueAddLong(MidiMessageQueue* queue, UBYTE* data, UINT32 size,
		      INT32 sysexIndex, INT64 timestamp, int overwrite);

// returns NULL if no messages in queue.
MidiMessage* MIDI_QueueRead(MidiMessageQueue* queue);
// message will be removed from queue.
void MIDI_QueueRemove(MidiMessageQueue* queue, INT32 onlyLocked);
void MIDI_QueueClear(MidiMessageQueue* queue);

#endif


/*
 * Platform MIDI IN support.
 * deviceId:		device-by-number
 * deviceHandle:	native device handle
 */

#if USE_PLATFORM_MIDI_IN == TRUE

// number of messages to be buffered
#define MIDI_IN_MESSAGE_QUEUE_SIZE 64

// number of sysex to be buffered
#define MIDI_IN_LONG_QUEUE_SIZE 20
// maximum number of bytes in one sys ex message
#define MIDI_IN_LONG_MESSAGE_SIZE 1024

/*
 * Get the number of MIDI IN devices on the system.
 */
INT32	MIDI_IN_GetNumDevices(void);

/*
 * Get the name of the device with this id
 * Returns TRUE (1) or FALSE (0)
 */
INT32	MIDI_IN_GetDeviceName(INT32 deviceID, char *name, UINT32 nameLength);

/*
 * Get the vendor of the device with this id
 * Returns TRUE (1) or FALSE (0)
 */
INT32	MIDI_IN_GetDeviceVendor(INT32 deviceID, char *name, UINT32 nameLength);

/*
 * Get the description of the device with this id
 * Returns TRUE (1) or FALSE (0)
 */
INT32	MIDI_IN_GetDeviceDescription(INT32 deviceID, char *name, UINT32 nameLength);

/*
 * Get the version of the device with this id
 * Returns TRUE (1) or FALSE (0)
 */
INT32	MIDI_IN_GetDeviceVersion(INT32 deviceID, char *name, UINT32 nameLength);

/*
 * Open the device with this id.
 * Returns a device handle or 0 for failure.
 * Returns TRUE (1) or FALSE (0)
 */
MidiDeviceHandle* MIDI_IN_OpenDevice(INT32 deviceID);

/*
 * Close the device handle.
 * Returns TRUE (1) or FALSE (0)
 */
INT32	MIDI_IN_CloseDevice(MidiDeviceHandle* handle);

/*
 * Start the device with this handle.
 * Returns TRUE (1) or FALSE (0)
 */
INT32	MIDI_IN_StartDevice(MidiDeviceHandle* handle);

/*
 * Stop the device with this handle.
 * Returns TRUE (1) or FALSE (0)
 */
INT32	MIDI_IN_StopDevice(MidiDeviceHandle* handle);

/*
 * Get the next message from the queue.
 * This call blocks until the device is stopped
 * or a message is received.
 * The returned message is READ ONLY.
 * The message must be returned into the message
 * queue by calling MIDI_IN_ReleaseMessage.
 */
MidiMessage* MIDI_IN_GetMessage(MidiDeviceHandle* handle);

/*
 * Put a message, which was taken
 * out of the queue, back into the queue.
 */
void MIDI_IN_ReleaseMessage(MidiDeviceHandle* handle, MidiMessage* msg);

#endif // USE_PLATFORM_MIDI_IN


/*
 * Platform MIDI OUT support.
 * deviceId:		device-by-number
 * deviceHandle:	native device handle
 */

#if USE_PLATFORM_MIDI_OUT == TRUE

// number of sysex to be buffered
#define MIDI_OUT_LONG_QUEUE_SIZE 20

/*
 * Get the number of MIDI OUT devices on the system.
 */
INT32	MIDI_OUT_GetNumDevices(void);

/*
 * Get the name of the device with this id
 */
INT32	MIDI_OUT_GetDeviceName(INT32 deviceID, char *name, UINT32 nameLength);

/*
 * Get the vendor of the device with this id
 */
INT32	MIDI_OUT_GetDeviceVendor(INT32 deviceID, char *name, UINT32 nameLength);

/*
 * Get the description of the device with this id
 */
INT32	MIDI_OUT_GetDeviceDescription(INT32 deviceID, char *name, UINT32 nameLength);

/*
 * Get the version of the device with this id
 */
INT32	MIDI_OUT_GetDeviceVersion(INT32 deviceID, char *name, UINT32 nameLength);

/*
 * Open the device with this id.
 * Returns a device handle or 0 for failure.
 */
MidiDeviceHandle* MIDI_OUT_OpenDevice(INT32 deviceID);

/*
 * Close the device handle.
 * Returns TRUE (1) or FALSE (0)
 */
INT32	MIDI_OUT_CloseDevice(MidiDeviceHandle* handle);

/*
 * Send a short message to the hardware.
 * packedMsg: (status | data1<<8 | data2<<16) in platform-endianness
 * Timestamp is in microseconds.
 * Returns TRUE (1) or FALSE (0)
 */
INT32	MIDI_OUT_SendShortMessage(MidiDeviceHandle* handle, UINT32 packedMsg, UINT32 timestamp);

/*
 * Send a long message to the hardware.  Timestamp is in microseconds.
 * This blocks until a slot to send a message is free.
 * Returns TRUE (1) or FALSE (0)
 */
INT32	MIDI_OUT_SendLongMessage(MidiDeviceHandle* handle, UBYTE* data, UINT32 size, UINT32 timestamp);

#endif // USE_PLATFORM_MIDI_OUT

#endif // PLATFORM_MIDI_INCLUDED
