/*
 * @(#)GlyphVector.h	1.22 03/01/23
 *
 * Copyright 2003 Sun Microsystems, Inc. All rights reserved.
 * SUN PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */

/*
 * @author Charlton Innovations, Inc.
 */

#include "FontGlue.h"
#include "TX.h"
#include "GeneralPath.h"
#include "GlyphImageRef.h"

// REMIND: speedup if we write a "hsDesc" which can live on the stack

class StrikeTable;

class GlyphVector {
private:

    fontObject * const fFont;
    JNIEnv * const fEnv;

    Boolean fNeedShaping;

    UInt32  fNumGlyphs;

    // default storage, to avoid heap allocations
    enum {
        eDefaultStorage = 100
    };
    UInt32   fBaseGlyphs[eDefaultStorage];
    UInt32   fBaseCharIndices[eDefaultStorage];
    jfloat   fBasePositions[eDefaultStorage*2];
    ImageRef fBaseImageRefs[eDefaultStorage];

    UInt32   *fGlyphs;
    UInt32   *fCharIndices;
    jfloat   *fPositions;
    ImageRef *fImageRefs;

    jdouble  *fTransforms;
    jint     *fTXIndices;
    jint      fTXCount;

    TX fFontTX; // font transform
    TX fDevTX; // device transform, assume will frequently be identity

    jboolean fIsAntiAliased;
    jboolean fUsesFractionalMetrics;

    void setNumGlyphs(jint);

    void allocateGlyphs(jint);
    void allocateCharIndices(void);
    void allocatePositions(void);
    void allocateImageRefs(void);

    void initGlyphs(const jchar *theChars, jint count);
    void initCharIndices();

    Strike& getGlyphStrike(TX* devTX) const;

private:
    GlyphVector(); // no implementation
    GlyphVector(GlyphVector&);  // no implementation
    GlyphVector& operator=(const GlyphVector&); // no implementation

    // heap allocation is possible but be careful to clean up

    friend class StrikeTable;

public:
    GlyphVector(JNIEnv *env,
                jdoubleArray fontTX, jdoubleArray devTX,
                jboolean isAntiAliased, jboolean usesFractionalMetrics,
                fontObject *fo);
    ~GlyphVector();

    /**
     * initialize glyph data from text or pure glyph codes.
     */
    void setString(jstring theString);
    void setText(jcharArray chars, jint offset, jint count);
    void setText(jbyteArray bytes, jint offset, jint count);
    void setGlyphCodes(jintArray glyphs, jint offset, jint count);
    void setGlyphCodes(jintArray glyphs);

    /**
     * Initialize position and transform data.
     */
    void setPositions(jfloatArray newPos);
    void setTransforms(jdoubleArray jTransforms, jintArray jTXIndices);

    /**
     * Return true if special layout is required.
     */
    inline Boolean needShaping() const { return fNeedShaping; }

    /**
     * Prepare blit info data.  positionAndGetImages will compute glyph positions
     * using default advances.  This is used for rendering Strings.  For
     * rendering java GlyphVector objects, which already have positions, use
     * getImages instead.
     *
     * x and y are in the device coordinate system.
     */
    void positionAndGetImages(jfloat x, jfloat y);
    void getImages(jfloat x, jfloat y);

    /**
     * Lay out the glyphvector.
     */
    void positionGlyphs();

    /**
     * Return the number of glyphs in the glyphvector.
     */
    inline unsigned int getNumGlyphs(void) const { return fNumGlyphs; }

    /**
     * Get information about the raster image for a glyph.
     */
    inline void getGlyphBlitInfo(
        int index,
        const void * &pixels,
        unsigned int &rowBytes,
	int &x,
	int &y,
        int &width,
        int &height) const
    {
        pixels      = fImageRefs[index].pixels;
        rowBytes    = fImageRefs[index].rowBytes;
	x           = fImageRefs[index].x;
	y           = fImageRefs[index].y;
        width       = fImageRefs[index].width;
        height      = fImageRefs[index].height;
    }

    /**
     * Get pointer to array of image refs.
     * This does not initialise the array and so should only be used
     * where the array is known to be already populated.
     * Care must also be taken to use this only in a context where the
     * images can be relied upon to remain in the cache.
     */
    inline struct ImageRef* getGlyphImages() {
	return fImageRefs;
    }

    void getGlyphCodes(jobject gvSource);
    void getPositions(jobject gvSource);
    void getTransforms(jobject gvSource);
    void getGlyphVector(jobject gvSource);

    void writeGlyphCodes(jobject gvTarget);
    void writeCharIndices(jobject gvTarget);
    void writePositions(jobject gvTarget);

    jobject getGlyphLogicalBounds(jint index);

    jobject getVisualBounds();
    jobject getGlyphVisualBounds(jint index);

    jobject getPixelBounds(jfloat x, jfloat y);
    jobject getGlyphPixelBounds(jint index, jfloat x, jfloat y);

    jobject getOutline(jfloat x, jfloat y);
    jobject getGlyphOutline(jint glyphIndex, jfloat x, jfloat y);
    jobject getGlyphsOutline(jint start, jint count, jfloat x, jfloat y);

    jfloatArray getGlyphInfo();

    static void scanGlyph(GeneralPath& gp, 
			  Strike& theStrike, 
			  hsPathSpline& path, 
			  CurveWalker& walker, 
			  float xPos, float yPos,
			  jint glyph);

    static jboolean charsToGlyphs(Strike& theStrike, const jchar* theChars, UInt32* theGlyphs, jint count);
};
