/*
 * @(#)TX.h	1.6 03/01/23
 *
 * Copyright 2003 Sun Microsystems, Inc. All rights reserved.
 * SUN PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */

#ifndef __TX_header
#define __TX_header

#include <jni.h>
#include <jni_util.h>

// simple affine transform for use in native

class TX {
public:
    jdouble m00; // scale x
    jdouble m10; // shear y
    jdouble m01; // shear x		      
    jdouble m11; // scale y
    jdouble m02; // trans x
    jdouble m12; // trans y
    jint    fType; // OR of trans, scale, shear masks

    // different types than Java
    enum {
      TYPE_IDENTITY = 0,
      MASK_TRANS = 1,
      MASK_SCALE = 2,
      MASK_SHEAR = 4
    };

    typedef enum {
      XAxis = 0,
      YAxis = 1
    } EAxis;

    TX() {
      setIdentity();
    }

    void setIdentity() {
      fType = TYPE_IDENTITY;
      m00 = 1;
      m10 = 0;
      m01 = 0;
      m11 = 1;
      m02 = 0;
      m12 = 0;
    }

    TX(double *matrix) {
      if (matrix) {
	set(matrix);
      } else {
	setIdentity();
      }
    }

    TX(const TX &rhs) 
	: m00(rhs.m00)
	, m10(rhs.m10)
	, m01(rhs.m01)
	, m11(rhs.m11)
	, m02(rhs.m02)
	, m12(rhs.m12)
	, fType(rhs.fType) {
    }

    void set(const TX& rhs) {
	m00 = rhs.m00;
	m10 = rhs.m10;
	m01 = rhs.m01;
	m11 = rhs.m11;
	m02 = rhs.m02;
	m12 = rhs.m12;
	fType = rhs.fType;
    }

    void set(jdouble* matrix) {
      if (matrix) {
	m00 = matrix[0];
	m10 = matrix[1];
	m01 = matrix[2];
	m11 = matrix[3];
	m02 = matrix[4];
	m12 = matrix[5];
	initType();
      }
    }

    void get(jdouble* matrix) const {
      if (matrix) {
	matrix[0] = m00;
	matrix[1] = m10;
	matrix[2] = m01;
	matrix[3] = m11;
	matrix[4] = m02;
	matrix[5] = m12;
      }
    }

    void initType() {
      fType = TYPE_IDENTITY;
      if (m00 != 1 || m11 != 1) {
	fType |= MASK_SCALE;
      }
      if (m01 != 0 || m10 != 0) {
	fType |= MASK_SHEAR;
      }
      if (m02 != 0 || m12 != 0) {
	fType |= MASK_TRANS;
      }
    }

    TX(JNIEnv *env, jdoubleArray matRef) {
      if (JNU_IsNull(env, matRef)) {
	setIdentity();
      } else {
        jint matSize = env->GetArrayLength(matRef);
        if (matSize < 4) {
	  setIdentity();
	} else {
	  jdouble matrix[6];
	  if (matSize < 6) {
	    env->GetDoubleArrayRegion(matRef, 0, 4, matrix);
	    matrix[4] = matrix[5] = 0;
	  } else {
	    env->GetDoubleArrayRegion(matRef, 0, 6, matrix);
	  }
	  set(matrix);
        }
      }
    }

    void concat(const TX& rhs);

    void transform(jfloat points[], int count) const {
      transform(points, count, fType);
    }

    void vectorTransform(jfloat points[], int count) const {
      transform(points, count, fType & ~MASK_TRANS);
    }

    void transform(jfloat &x, jfloat &y) const {
      transform(x, y, fType);
    }

    void vectorTransform(jfloat &x, jfloat &y) const {
      transform(x, y, fType & ~MASK_TRANS);
    }

    jboolean isIdentity() const {
      return fType == TYPE_IDENTITY;
    }

    jboolean isTranslation() const {
      return fType == MASK_TRANS;
    }

    jboolean rotatesAxis(EAxis axis) const {
      return ((axis == XAxis) ? m10 : m01) != 0;
    }

    jboolean invert();

private:
    void transform(jfloat points[], int count, int type) const;
    void transform(jfloat &x, jfloat &y, int type) const;
};

#endif // __TX_header
