/*
 * @(#)fontObject.h	1.49 03/01/23
 *
 * Copyright 2003 Sun Microsystems, Inc. All rights reserved.
 * SUN PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */

#ifndef fontObjectDefined
#define fontObjectDefined

#include <assert.h>
#include <string.h>
#include <stdlib.h>

#include <jni.h>

#include "HeadSpin.h"

extern int debugFonts; // set by env. JAVA2D_DEBUGFONTS
class Strike;           // forward referenced
class FontTransform;    // forward referenced

typedef unsigned char byte;

class fontObject;

typedef UInt16 Unicode;		// FIXME: This definition should probably be removed in favor of Unicode16
typedef UInt16 Unicode16;
typedef UInt32 Unicode32;

enum FontFormats {
    kTrueTypeFontFormat,
    kType1FontFormat,
    kT2KFontFormat,
    kTTCFontFormat,
    kCompositeFontFormat,
    kNRFontFormat,
    kInvalidFontFormat = -1
};

#define kFullName 4
#define kFamilyName 1
#define kPSFontName 6

#ifdef __cplusplus

typedef struct {
    Boolean multipleMasterFont;
    char *fontName;
    char *fullName;
    char *familyName;
} t1FontInfo;

typedef struct {

  char   *fontName;
  char   *fullName;
  char   *familyName;

} fontNameRecord;

typedef struct {
    int fd;
    unsigned char *buffer;
    size_t buflen;
    size_t boffs;
    unsigned char ch;
    int eof;
    int curToken;
    int prevToken;
} sDataRecord;

class CharToGlyphMapper {
public:
    virtual ~CharToGlyphMapper() {}

    virtual void CharsToGlyphs(
        int count, const Unicode16 unicodes[], UInt32 glyphs[]) const = 0;

    virtual void CharsToGlyphs(
        int count, const Unicode32 unicodes[], UInt32 glyphs[]) const = 0;

    virtual Boolean canDisplay(Unicode ch) const = 0;

    virtual int getMissingGlyphCode() const { return 0; }

    UInt32 CharToGlyph(const Unicode16 unicode) const {
        UInt32 glyphCode;
        CharsToGlyphs(1, &unicode, &glyphCode);
        return glyphCode;
    }

    UInt32 CharToGlyph(const Unicode32 unicode) const {
        UInt32 glyphCode;
        CharsToGlyphs(1, &unicode, &glyphCode);
        return glyphCode;
    }
};

struct FontStrikeDesc {
    fontObject      *fFontObj;
    float           fMatrix[4];
    Int32           fDoAntiAlias;
    Int32           fDoFracEnable;
    int             fStyle;
    int             fIndex; 
};
#define kFontStrikeDescTag  0x46736474   // atol("Fsdt")

class fontObject {

private:
    Strike *fStrike;        // cached for getStrike

    // cached state for getStrike
    FontTransform *fStrikeTx;
    jboolean fStrikeIsAntiAliased;
    jboolean fStrikeUsesFractionalMetrics;
    int fStrikeCurrentStyle;

public:
    enum FontFormats fFormat;
    Unicode *fFontname;
    int  fFontNameLen;
    Unicode * fFullName;             /* Full font name (english)   */
    Unicode * fFamilyName;           /* Family font name (english) */
    Unicode * fFullNameAlt;          /* Full font name (native)    */
    Unicode * fFamilyNameAlt;        /* Family font name (native)  */
    Unicode * fPSFontName;           /* Postscript name */
    // surrogates won't fit in here, but this code is needed only to
    // get back a char from a glyph so the NR can be used, and there's
    // no case in which the NR is used there. So if a glyph isn't found
    // in here we back off to T2K which is what would probably happen anyway.
    // This array is allocated in the CMapper to hold numGlyphs and is
    // dynamically populated as chars are mapped to glyphs.
    // Not used for composites, natives. Just for TT & Type1
    Unicode16* fGlyphToCharMap;
    int  fGlyphCharMapLen;

    void setStrike( Strike *theStrike, FontTransform& tx,
                    jboolean isAntiAliased,
                    jboolean usesFractionalMetrics);

 public:
    int fCurFont;                       //- current logical font
    int fFontCount;                     //- number of logical sub-fonts

    Boolean isValid;                    //- For lazy font verification; if false, font is not usable
    									  
    Boolean fFontNamesValid;
    Boolean fUseT2KRasterizer;

    fontNameRecord theFontNames;
    // transient ... REMIND: change communication between FontWrapper & Strike
    int m_currentStyle;
    int fUnitsPerEM;
    int fMacStyle;
	int fRealStyle;
    int fisISFOCFont;

    fontObject() :
        fStrike(NULL),
        fStrikeTx(NULL),
        fStrikeIsAntiAliased(false),
        fStrikeUsesFractionalMetrics(false),
        fStrikeCurrentStyle(0),
        fFormat(kInvalidFontFormat),
        fFontname(0),
        fFontNameLen(0),
        fFullName(0),
        fFamilyName(0),
        fFullNameAlt(0),
        fFamilyNameAlt(0),
	fPSFontName(0),
	fGlyphToCharMap(NULL),
	fGlyphCharMapLen(0),
        fCurFont(0),
        fFontCount(0),
        isValid(true),
        fFontNamesValid(false),
	fUseT2KRasterizer(false),
        m_currentStyle(0),
        fUnitsPerEM(0),
        fMacStyle(0),
		fRealStyle(0),
        fisISFOCFont(0)
    {
    }

    virtual ~fontObject();

    virtual const void* ReadChunk(UInt32 offset,
                                  UInt32 length, void* chunk = nil) = 0;
    virtual void readBlock(size_t offs, size_t size, char *dest) = 0;
    virtual void ReleaseChunk(const void* chunk) = 0;
    virtual void FileClose() {};
    virtual Boolean MatchName(int nameID, const UInt16 *name, int nameLen) = 0;
    virtual int GetName(UInt16& platformID, UInt16& scriptID, UInt16& languageID, UInt16& nameID, UInt16 *name) = 0;
    virtual const Unicode *GetFontName(int& nameLen) = 0;
    virtual const char *GetFontLocalPlatformName() const { return NULL;}
    virtual const char *GetFontNativeName (int index=0) const { return NULL;}
    virtual const Unicode *GetFileName(int& nameLen) const { nameLen = 0; return NULL; }
    virtual const int GetNumberOfNativeNames() const { return 0; }

    virtual const UInt16 GetUnitsPerEM() { return fUnitsPerEM;}
    virtual const UInt16 GetMacStyle()   { return fMacStyle;}
	virtual const UInt16 GetRealStyle()  { return fRealStyle;}

    enum FontFormats GetFormat() const { return fFormat; }

    //  These are for TrueType
    virtual UInt32 FindTableSize(UInt32 tableTag, UInt32* offset);
    const void* ReadTable(UInt32 tableTag, UInt32* tableLength = nil);
    virtual const void * GetTable(UInt32 tableTag) { return (void *)0; }

    //  These are for T2K
    virtual const void* GetData() { return this->ReadChunk(0, 0x7fffffff); }
    virtual UInt32 GetDataLength() = 0;

    // informational methods
    virtual enum FontFormats getScalerID(
                                FontTransform& tx,
                                jboolean isAntiAliased,
                                jboolean usesFractionalMetrics)
    {
        return kT2KFontFormat;
    }

    Strike& getStrike(  FontTransform& tx,
                        jboolean isAntiAliased,
                        jboolean usesFractionalMetrics);
    Strike& getStrike();
};

class fileFontObject;

void fontfilecachecheck(fileFontObject * fo);
void fontfilecachedelete(fileFontObject * fo);

// Classes derived from fontObject
class fileFontObject : public fontObject {
 private:
    int fFontNativeNameCount;
    // can be multibyte in locale
    char **fFontNativeNames;
    char *fFontLocalPlatformName;

    // always UNICODE
    Unicode * fFileName;
    int fFileNameLen;
    int fUseCount;
    int fFile;
    UInt32 fFileSize;

 public:
    fileFontObject() :
	fFontNativeNameCount(0),
        fFontNativeNames(0),
        fFontLocalPlatformName(0),
	
        fFileName(0),
        fFileNameLen(0),
        fUseCount(0),
        fFile(-1),
        fFileSize(0) {}

    ~fileFontObject() {
#ifndef CLOSE_ON_USECOUNT
	fontfilecachedelete(this);
#endif
	if (fFileName != NULL) {
	    delete[] fFileName;
	    fFileName = NULL;
	}
	if (fFontNativeNames != NULL) {
	    for (int i=0; i<fFontNativeNameCount; i++) {
		if (fFontNativeNames[i] != NULL) {
		    free((void *)fFontNativeNames[i]);
		    fFontNativeNames[i] = NULL;
		}
	    }
	    delete[] fFontNativeNames;
	    fFontNativeNames = NULL;
	}	    
	if (fFontLocalPlatformName != NULL) {
	     free((void *)fFontLocalPlatformName);
	    fFontLocalPlatformName = NULL;
	}
        assert(fUseCount == 0);
    }


    virtual const void* ReadChunk(UInt32 offset, UInt32 length,
                                  void* chunk = nil);

    virtual void readBlock(size_t offset, size_t size, char *dest);

    virtual void ReleaseChunk(const void* chunk);

    virtual void FileClose();

    virtual UInt32 GetDataLength();

    virtual const UInt16 GetUnitsPerEM();

    virtual Boolean Init(const Unicode *fileName, int nameLen,
                         const char *localPlatformName,
                         const char *nativeName,
                         enum FontFormats format, int logicalNumber = 0) {
        fFileName = new Unicode[nameLen];
        fFileNameLen = nameLen;
        memcpy(fFileName, fileName, nameLen * sizeof(Unicode));
	fFontNativeNames = new char*[1];
	fFontLocalPlatformName = strdup(localPlatformName);
	if (nativeName != NULL) {
	    fFontNativeNames[0] = strdup(nativeName);
	    fFontNativeNameCount = 1;
	}
        fFormat = format;
        return true;
    }
    virtual const Unicode * GetFileName(int& nameLen) const {
	nameLen = fFileNameLen; return fFileName;
    }
    virtual const char * GetFontLocalPlatformName() const {
	return fFontLocalPlatformName;
    }
    virtual const char * GetFontNativeName(int index = 0) const {
	if (fFontNativeNameCount > index) {
	    return fFontNativeNames[index] ;
	} else {
	    return NULL;
	}
    }
    virtual const int GetNumberOfNativeNames() const {
	return fFontNativeNameCount;
    }
 
    virtual void AddNativeName(const char *nativeName) {
        int i=0;
	if (nativeName != NULL) {
	    for (i=0; i< fFontNativeNameCount; i++) {
		if (strcmp(nativeName,fFontNativeNames[i]) == 0) {
		    return;
		}
	    }
	    char **nnames = new char*[fFontNativeNameCount+1];
	    for (i=0; i<fFontNativeNameCount; i++) {
		nnames[i] = fFontNativeNames[i];
	    }
	    nnames[fFontNativeNameCount] = strdup(nativeName);
	    delete fFontNativeNames;
	    fFontNativeNames = nnames;
	    fFontNativeNameCount++;
	}
    }
};

class sfntFileFontObject : public fileFontObject {
  protected:
    UInt32 *offsets;
    char*  fGposTable;
    char*  fGdefTable;
    char*  fGsubTable;
    char*  fMortTable;
  public:
    sfntFileFontObject() : 
        offsets(0), 
        fGposTable(0), 
        fGdefTable(0), 
        fGsubTable(0), 
        fMortTable(0),
        fFontNameTable(NULL),
        fHintsTableChecked(false),
        fContainsHintsTable(false),
        fCmap(NULL) {};
    virtual ~sfntFileFontObject();

    virtual Boolean Init(const Unicode *fileName, int nameLen, 
			 const char *localPlatformName,
			 const char *nativeName,
                         enum FontFormats format, int logicalNumber = 0);
    virtual Boolean MatchName(int nameID, const UInt16 *name, int nameLen);
    virtual int GetName(UInt16& platformID, UInt16& scriptID, UInt16& languageID, UInt16& nameID, UInt16 *name);
    virtual void ReadNames();
    virtual const Unicode *GetFontName(int& nameLen);
    virtual UInt32 FindTableSize(UInt32 tableTag, UInt32* offset);
    virtual const void * GetTable(UInt32 tableTag);
    inline int getLogicalFontCount() {return fFontCount;};
    // informational methods
    virtual enum FontFormats getScalerID(
                                FontTransform& tx,
                                jboolean isAntiAliased,
                                jboolean usesFractionalMetrics);
    char *fFontNameTable;
    Boolean fHintsTableChecked;
    Boolean fContainsHintsTable;
    byte  *fCmap;
};

class type1FileFontObject : public fileFontObject {
public:
    virtual Boolean MatchName(int nameID, const UInt16 *name, int nameLen);
    virtual int GetName(UInt16& platformID, UInt16& scriptID, UInt16& languageID, UInt16& nameID, UInt16 *name);
    virtual const Unicode *GetFontName(int& nameLen);
    void DeleteFontInfo(t1FontInfo *fontInfo);
    t1FontInfo* GetFontInfo();
    virtual enum FontFormats getScalerID(
                                FontTransform& tx,
                                jboolean isAntiAliased,
                                jboolean usesFractionalMetrics);


private:
    int OpenFont(const char *fileName, sDataRecord *sData);
    void CloseFont(sDataRecord *sData);
    int NextToken(sDataRecord *sData, char *token, int toklen);
    int SkipSpaces(sDataRecord *sData);
    int NextChar(sDataRecord *sData);
};

void registerFont(enum FontFormats format, fontObject *cf);
fontObject *GetFontObject(const Unicode *fileName, int nameLength);
fontObject *GetFontObject(unsigned index);

/* Needed for native rasterizer support */
fileFontObject *GetNativeFontObject ();
void MakeNativeMapper (const char *fontString);


#endif

#ifdef __cplusplus
extern "C" {
#endif

#ifdef WIN32
sfntFileFontObject*  getLinkedEUDCFont();
#endif

    int CountFontObjects();

    // Finds a font object that has the given name as a family name or face name
    // and supports the given style (more or less).
    // Returns NULL if no such font object has been registered.
    fontObject *FindFontObject(const UInt16 * name, unsigned int nameLen, int style);

    // Returns a font object of last resort.
    fontObject *DefaultFontObject();

    const void *FOReadChunk(fontObject *fo, UInt32 offset, UInt32 length);
    void FOReleaseChunk(fontObject *fo, const void *chunk);
    long FOFindTableSize(fontObject *fo, UInt32 tableTag, UInt32 *offset);

    // These are need by T2K
    const void* FOGetData(fontObject *fo);      // return the entire font
    UInt32 FOGetDataLength(fontObject *fo);
    const Unicode *CreateTrueTypeFont ( Unicode *fontFileName, int fontNameLen, char *localName, int *nameLen );
    JNIEXPORT jboolean JNICALL AWTIsHeadless();
    jboolean FOFontSupportsEncoding(JNIEnv *env, jobject theFont, jstring theEncoding);
#ifdef __cplusplus
}
#endif

#endif
