/*
 * @(#)MemCache.h	1.3 03/01/23
 *
 * Copyright 2003 Sun Microsystems, Inc. All rights reserved.
 * SUN PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */

#ifndef __MEMCACHE_H__
#define __MEMCACHE_H__

#include <assert.h>
#include <HeadSpin.h>

struct MemEntry;

class MemCache {
  UInt16 fNumEntries;   // number of entries currently in cache
  UInt16 fMaxEntries;   // maximum entries in cache
  UInt32 fMem;          // amount of memory currently used by cache
  UInt32 fMaxMem;       // maximum memory to be used by cache (can be exceeded)
  UInt16 fNearLimit;    // number of entries considered to be near the top
  UInt16 fTop;          // current top
  MemEntry* fEntries;   // the buffer of entries
  Boolean fPurge;       // if true, purge cache if needed, if false only purge on request.

public:
  /** 
   * Construct a cache that can hold numEntries glyphs, with a maximum memory
   * usage of maxMem, moving entries to the top of the cache if one is 
   * located further than nearLimit from the top of the cache.  If purgeOn is
   * true, enable automatic cache management.  If it is false, entries are
   * only flushed when all glyph entries are in use, and aging entries are
   * not refetched to the top of the queue when requested.
   *
   * MaxEntries must be < 0x7fffu.
   */
  MemCache(UInt16 maxEntries, UInt16 nearLimit, UInt32 maxMem, Boolean purgeOn);

  /**
   * Free the cache.
   */
  virtual ~MemCache();

  /**
   * The index for requests where no previous index exists.
   */
  static const UInt16 kNoIndex;

  /**
   * Turn cache management on or off.  When it is off, nothing gets purged from
   * the cache except by explicit request or when all glyph entries are in use.
   */
  void setPurgeOn(Boolean onOff) {
    fPurge = onOff;
  }

  /**
   * Reserve space for numEntries glyphs and memSize memory in the cache.  This
   * will purge glyphs if required to free up the necessary space.  If purged
   * is not null, it will return true if glyphs were purged.
   */
  void reserveSpace(UInt16 numEntries, UInt32 memSize, Boolean* purged);

  /**
   * Assign new memory of size bytes to cache, and associate with the 
   * ownerID.  The cache index is returned.  The memory can later be
   * freed using HSMemory::Delete.
   *
   * This may cause purging of memory in the cache.  If purged is not
   * null, the value it points to will be set to true if memory was purged.
   *
   * The memory is only guaranteed available until the next call to
   * assignMemory or getMemory. 
   *
   * OwnerID may be any 32-bit value.
   * Size must be < 0x7fffffffu.
   */
   UInt16 assignMemory(UInt32 ownerID, UInt32 size, void* memory, Boolean* purged = 0);

  /**
   * Return a pointer to the previously allocated memory associated with 
   * the ownerID and index.  The pointer can be null if the cache no 
   * longer holds memory for this owner at this index.
   *
   * This may cause purging of memory associated with other indexes.  If
   * purged is not null, the value it points to will be set to true if
   * memory was purged.
   *
   * The memory is only guaranteed available until the next call to
   * assignMemory or getMemory. 
   */
  const void* getMemory(UInt32 ownerID, UInt16 &index, Boolean* purged = 0);

  /**
   * Return a pointer to the previously allocated memory associated with
   * the ownerID and index.  The pointer can be null if the cache no
   * longer holds memory for this owner at this index.
   *
   * This will not cause purging of memory in the cache.
   *
   * The memory is only guaranteed available until the next call to
   * assignMemory or getMemory. 
   */
  void* refMemory(UInt32 ownerID, UInt16 index);

  /**
   * Free the memory at index if it belongs to owner.
   */
  void releaseMemory(UInt32 ownerID, UInt16 index, Boolean* purged = 0);

  /**
   * Free all memory held by the cache.
   */
  void releaseAll();

protected:
  virtual void disposeMemPtr(void* mem) = 0;

private:
  UInt16 next(UInt16 index) const { 
    if (++index >= fMaxEntries) {
      index = 0; 
    }
    return index; 
  }

  void releaseEntry(MemEntry& e, Boolean* purged);

  void setEntry(MemEntry& e, UInt32 ownerID, void* memPtr, UInt32 memSize, Boolean* purged);

  void moveEntry(MemEntry& lhs, MemEntry& rhs, Boolean* purged);
};

// __MEMCACHE_H__
#endif
