/*
 * @(#)Win32GraphicsConfig.java	1.37 03/01/23
 *
 * Copyright 2003 Sun Microsystems, Inc. All rights reserved.
 * SUN PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */

package sun.awt;

import java.awt.BufferCapabilities;
import java.awt.GraphicsConfiguration;
import java.awt.GraphicsDevice;
import java.awt.ImageCapabilities;
import java.awt.Toolkit;
import java.awt.Transparency;
import java.awt.image.BufferedImage;
import java.awt.image.ColorModel;
import java.awt.image.DirectColorModel;
import java.awt.image.Raster;
import java.awt.image.VolatileImage;
import java.awt.image.WritableRaster;
import java.awt.geom.AffineTransform;
import java.awt.Rectangle;

import java.awt.color.ColorSpace;
import java.awt.image.DataBuffer;

import sun.java2d.SurfaceData;
import sun.java2d.loops.RenderLoops;
import sun.java2d.loops.SurfaceType;
import sun.java2d.loops.CompositeType;
import sun.awt.windows.WVolatileImage;
import sun.awt.windows.Win32PeerlessImage;
import sun.awt.windows.Win32OffScreenSurfaceData;

/**
 * This is an implementation of a GraphicsConfiguration object for a
 * single Win32 visual.
 *
 * @see GraphicsEnvironment
 * @see GraphicsDevice
 * @version 10 Feb 1997
 */
public class Win32GraphicsConfig extends GraphicsConfiguration {
    Win32GraphicsDevice screen;
    int visual;  //PixelFormatID
    RenderLoops solidloops;
    private static BufferCapabilities bufferCaps;
    private static ImageCapabilities imageCaps;    

	private static native void initIDs();

	static {
		initIDs();
	}
	
	/**
	 * Returns a Win32GraphicsConfiguration object with the given device
	 * and PixelFormat.  Note that this method does NOT check to ensure that
	 * the returned Win32GraphicsConfig will correctly support rendering into a
	 * Java window.  This method is provided so that client code can do its
	 * own checking as to the appropriateness of a particular PixelFormat.  
	 * Safer access to Win32GraphicsConfigurations is provided by 
	 * Win32GraphicsDevice.getConfigurations().
	 */
	public static Win32GraphicsConfig getConfig(Win32GraphicsDevice device,
	 int pixFormatID) {
		return new Win32GraphicsConfig(device, pixFormatID);	
	}

    /**
     * @deprecated as of JDK version 1.3
     * replaced by <code>getConfig()</code>
     */
    public Win32GraphicsConfig(GraphicsDevice device, int visualnum) {
	this.screen = (Win32GraphicsDevice)device;
	this.visual = visualnum;
    }

    /**
     * Return the graphics device associated with this configuration.
     */
    public GraphicsDevice getDevice() {
	return screen;
    }

	/**
	 * Return the PixelFormatIndex this GraphicsConfig uses
	 */
	public int getVisual() {
	return visual;
	}

    /**
     * Return the RenderLoops this type of destination uses for
     * solid fills and strokes.
     */
    public synchronized RenderLoops getSolidLoops(SurfaceType stype) {
	if (solidloops == null) {
	    solidloops = SurfaceData.makeRenderLoops(SurfaceType.OpaqueColor,
						     CompositeType.SrcNoEa,
						     stype);
	}
	return solidloops;
    }

    /**
     * Returns a BufferedImage with raster and color model
     * compatible with this graphics configuration.  This method
     * has nothing to do with memory-mapping
     * a device.  This BufferedImage has
     * a raster and color model
     * that is closest to this native device configuration and thus
     * can be optimally blitted to this device.
     */
    public BufferedImage createCompatibleImage(int width, int height) {
	// REMIND: need to take into account the visual type
        int numpixels = width * height;
	ColorModel model = getColorModel();
	WritableRaster raster =
            model.createCompatibleWritableRaster(width, height);
        
        
        if (Win32PeerlessImage.isValidOpaqueConfig(model, model, numpixels)) {
            return new Win32PeerlessImage(model, model,
                                          raster, model.isAlphaPremultiplied(),
                                          Transparency.OPAQUE);
        } else {
            return new BufferedImage(model, raster,
                                     model.isAlphaPremultiplied(), null);
        }                                     
    }

    /**
     * Returns a VolatileImage
     * compatible with this graphics configuration.
     * The returned <code>VolatileImage</code> has
     * a layout and color model that is closest to this native device
     * and may have data that is stored on 
     * the device (i.e., in VRAM) and can therefore be rendered to and 
     * blitted from using hardware-acceleration techniques (i.e., DirectX).
     */
    public VolatileImage createCompatibleVolatileImage(int width, int height) {
	return (VolatileImage)(new WVolatileImage(this, width, height));
    }

    /**
     * Returns a BufferedImage that supports the specified transparency
     * and has a channel layout and color model
     * compatible with this graphics configuration.  This method
     * has nothing to do with memory-mapping
     * a device. This BufferedImage has a layout and
     * color model that can be optimally blitted to a device
     * with this configuration.
     * @see Transparency#OPAQUE
     * @see Transparency#BITMASK
     * @see Transparency#TRANSLUCENT
     */
    public BufferedImage createCompatibleImage(int width, int height,
					       int transparency) {
	// REMIND: need to take into account the visual type
        int numpixels = width * height;
        
        switch (transparency) {
        case Transparency.OPAQUE:
	    return createCompatibleImage(width, height);
            
        case Transparency.BITMASK:
        {
            ColorModel cmSw = getColorModel(transparency);
            ColorModel cmHw = getColorModel();
	    WritableRaster wr = cmSw.createCompatibleWritableRaster(width,
                                                                    height);

            if (Win32PeerlessImage.isValidBitmaskConfig(cmSw, cmHw, 
                                                        numpixels)) 
            {
                return new Win32PeerlessImage(cmSw, cmHw, wr, 
                                              cmSw.isAlphaPremultiplied(), 
                                              Transparency.BITMASK);
            } else {
                return new BufferedImage(cmSw, wr, 
                                         cmSw.isAlphaPremultiplied(), null);
            }
        }

        case Transparency.TRANSLUCENT:
        {
            ColorModel cmSw = getColorModel(transparency);
            ColorModel cmHw = cmSw;
	    WritableRaster wr = cmSw.createCompatibleWritableRaster(width,
                                                                    height);

	    if (Win32PeerlessImage.isValidTranslucentConfig(cmSw, cmHw, 
							    numpixels)) 
            {
                return new Win32PeerlessImage(cmSw, cmHw, wr, 
                                              cmSw.isAlphaPremultiplied(), 
                                              Transparency.TRANSLUCENT);
            } else {
		return new BufferedImage(cmSw, wr, 
                                         cmSw.isAlphaPremultiplied(), null);
            }
        }
    
        default:
            throw new IllegalArgumentException("Unknown transparency type "+
                                               transparency);
        }
    }

    /**
     * Returns the color model associated with this configuration.
     */
    public synchronized ColorModel getColorModel() {
	return screen.getColorModel();
    }

    /**
     * Returns a new color model for this configuration.  This call 
     * is only used internally, by images and components that are
     * associated with the graphics device.  When attributes of that
     * device change (for example, when the device palette is updated),
     * then this device-based color model will be updated internally
     * to reflect the new situation.
     */
    public ColorModel getDeviceColorModel() {
	return screen.getDynamicColorModel();
    }

    /**
     * Returns the color model associated with this configuration that
     * supports the specified transparency.
     */
    public ColorModel getColorModel(int transparency) {
        if (transparency == Transparency.OPAQUE) {
            return getColorModel();
        }
        else if (transparency == Transparency.BITMASK) {
            return new DirectColorModel(25, 0xff0000, 0xff00, 0xff, 0x1000000);
        }
        else if (transparency == Transparency.TRANSLUCENT) {
	    return getTranslucentColorModel();
        } else {
            return null;
        }
    }

    private static final int DCM_4444_RED_MASK = 0x0f00;
    private static final int DCM_4444_GRN_MASK = 0x00f0;
    private static final int DCM_4444_BLU_MASK = 0x000f;
    private static final int DCM_4444_ALP_MASK = 0xf000;
    static ColorModel translucentCM = null;
    public static ColorModel getTranslucentColorModel() {
	if (Win32OffScreenSurfaceData.getTextureBpp() == 16) {
	    if (translucentCM == null) {
		translucentCM = new DirectColorModel(16,
						     DCM_4444_RED_MASK,
						     DCM_4444_GRN_MASK,
						     DCM_4444_BLU_MASK,
						     DCM_4444_ALP_MASK);
	    }
	    return translucentCM;
	} else {
	    return ColorModel.getRGBdefault();
	}
    }

    /**
     * Returns the default Transform for this configuration.  This
     * Transform is typically the Identity transform for most normal
     * screens.  Device coordinates for screen and printer devices will
     * have the origin in the upper left-hand corner of the target region of
     * the device, with X coordinates
     * increasing to the right and Y coordinates increasing downwards.
     * For image buffers, this Transform will be the Identity transform.
     */
    public AffineTransform getDefaultTransform() {
	return new AffineTransform();
    }

    /**
     *
     * Returns a Transform that can be composed with the default Transform
     * of a Graphics2D so that 72 units in user space will equal 1 inch
     * in device space.
     * Given a Graphics2D, g, one can reset the transformation to create
     * such a mapping by using the following pseudocode:
     * <pre>
     *      GraphicsConfiguration gc = g.getGraphicsConfiguration();
     *
     *      g.setTransform(gc.getDefaultTransform());
     *      g.transform(gc.getNormalizingTransform());
     * </pre>
     * Note that sometimes this Transform will be identity (e.g. for
     * printers or metafile output) and that this Transform is only
     * as accurate as the information supplied by the underlying system.
     * For image buffers, this Transform will be the Identity transform,
     * since there is no valid distance measurement.
     */
    public AffineTransform getNormalizingTransform() {
	double xscale = getXResolution() / 72.0;
	double yscale = getYResolution() / 72.0;
	return new AffineTransform(xscale, 0.0, 0.0, yscale, 0.0, 0.0);
    }

    private double getXResolution() {
        return 72.0;
    }
    private double getYResolution() {
        return 72.0;
    }

    public String toString() {
	return ("Win32GraphicsConfig[dev="+screen+",pixfmt="+visual+"]");
    }

    private native Rectangle getBounds(int screen);
    
    public Rectangle getBounds() {
        return getBounds(screen.getScreen());
    }
    private static class DDrawBufferCapabilities extends BufferCapabilities {
        public DDrawBufferCapabilities(ImageCapabilities imageCaps) {
            super(imageCaps, imageCaps, FlipContents.PRIOR);
        }
        public boolean isFullScreenRequired() { return true; }
        public boolean isMultiBufferAvailable() { return true; }
    }
    
    private static class DDrawImageCapabilities extends ImageCapabilities {
        public DDrawImageCapabilities() {
            super(true);
        }
        public boolean isTrueVolatile() { return true; }
    }
    
    public BufferCapabilities getBufferCapabilities() {
        if (bufferCaps == null) {
            if (screen.isDirectDrawEnabled()) {
                bufferCaps = new DDrawBufferCapabilities(
                    getImageCapabilities());
            } else {
                bufferCaps = super.getBufferCapabilities();
            }
        }
        return bufferCaps;
    }
    
    public ImageCapabilities getImageCapabilities() {
        if (imageCaps == null) {
            if (screen.isDirectDrawEnabled()) {
                imageCaps = new DDrawImageCapabilities();
            } else {
                imageCaps = super.getImageCapabilities();
            }
        }
        return imageCaps;
    }
}
